@@ -1,10 +1,11 @@
 // Copyright (c) 2017-present PyO3 Project and Contributors
 
 use crate::attributes::TextSignatureAttribute;
+use crate::deprecations::Deprecation;
 use crate::params::{accept_args_kwargs, impl_arg_params};
 use crate::pyfunction::PyFunctionOptions;
 use crate::pyfunction::{PyFunctionArgPyO3Attributes, PyFunctionSignature};
-use crate::utils;
+use crate::utils::{self, PythonDoc};
 use crate::{deprecations::Deprecations, pyfunction::Argument};
 use proc_macro2::{Span, TokenStream};
 use quote::ToTokens;
@@ -65,8 +66,6 @@ impl<'a> FnArg<'a> {
 pub enum MethodTypeAttribute {
     /// #[new]
     New,
-    /// #[call]
-    Call,
     /// #[classmethod]
     ClassMethod,
     /// #[classattr]
@@ -84,7 +83,6 @@ pub enum FnType {
     Getter(SelfType),
     Setter(SelfType),
     Fn(SelfType),
-    FnCall(SelfType),
     FnNew,
     FnClass,
     FnStatic,
@@ -93,22 +91,27 @@ pub enum FnType {
 }
 
 impl FnType {
-    pub fn self_conversion(&self, cls: Option<&syn::Type>) -> TokenStream {
+    pub fn self_conversion(
+        &self,
+        cls: Option<&syn::Type>,
+        error_mode: ExtractErrorMode,
+    ) -> TokenStream {
         match self {
-            FnType::Getter(st) | FnType::Setter(st) | FnType::Fn(st) | FnType::FnCall(st) => {
-                st.receiver(cls.expect("no class given for Fn with a \"self\" receiver"))
-            }
+            FnType::Getter(st) | FnType::Setter(st) | FnType::Fn(st) => st.receiver(
+                cls.expect("no class given for Fn with a \"self\" receiver"),
+                error_mode,
+            ),
             FnType::FnNew | FnType::FnStatic | FnType::ClassAttribute => {
                 quote!()
             }
             FnType::FnClass => {
                 quote! {
-                    let _slf = pyo3::types::PyType::from_type_ptr(_py, _slf as *mut pyo3::ffi::PyTypeObject);
+                    let _slf = ::pyo3::types::PyType::from_type_ptr(_py, _slf as *mut ::pyo3::ffi::PyTypeObject);
                 }
             }
             FnType::FnModule => {
                 quote! {
-                    let _slf = _py.from_borrowed_ptr::<pyo3::types::PyModule>(_slf);
+                    let _slf = _py.from_borrowed_ptr::<::pyo3::types::PyModule>(_slf);
                 }
             }
         }
@@ -128,28 +131,47 @@ pub enum SelfType {
     TryFromPyCell(Span),
 }
 
+#[derive(Clone, Copy)]
+pub enum ExtractErrorMode {
+    NotImplemented,
+    Raise,
+}
+
 impl SelfType {
-    pub fn receiver(&self, cls: &syn::Type) -> TokenStream {
+    pub fn receiver(&self, cls: &syn::Type, error_mode: ExtractErrorMode) -> TokenStream {
+        let cell = match error_mode {
+            ExtractErrorMode::Raise => {
+                quote! { _py.from_borrowed_ptr::<::pyo3::PyAny>(_slf).downcast::<::pyo3::PyCell<#cls>>()? }
+            }
+            ExtractErrorMode::NotImplemented => {
+                quote! {
+                    match _py.from_borrowed_ptr::<::pyo3::PyAny>(_slf).downcast::<::pyo3::PyCell<#cls>>() {
+                        ::std::result::Result::Ok(cell) => cell,
+                        ::std::result::Result::Err(_) => return ::pyo3::callback::convert(_py, _py.NotImplemented()),
+                    }
+                }
+            }
+        };
         match self {
             SelfType::Receiver { mutable: false } => {
                 quote! {
-                    let _cell = _py.from_borrowed_ptr::<::pyo3::PyCell<#cls>>(_slf);
+                    let _cell = #cell;
                     let _ref = _cell.try_borrow()?;
                     let _slf = &_ref;
                 }
             }
             SelfType::Receiver { mutable: true } => {
                 quote! {
-                    let _cell = _py.from_borrowed_ptr::<::pyo3::PyCell<#cls>>(_slf);
+                    let _cell = #cell;
                     let mut _ref = _cell.try_borrow_mut()?;
                     let _slf = &mut _ref;
                 }
             }
             SelfType::TryFromPyCell(span) => {
                 quote_spanned! { *span =>
-                    let _cell = _py.from_borrowed_ptr::<::pyo3::PyCell<#cls>>(_slf);
+                    let _cell = #cell;
                     #[allow(clippy::useless_conversion)]  // In case _slf is PyCell<Self>
-                    let _slf = std::convert::TryFrom::try_from(_cell)?;
+                    let _slf = ::std::convert::TryFrom::try_from(_cell)?;
                 }
             }
         }
@@ -202,9 +224,10 @@ pub struct FnSpec<'a> {
     pub attrs: Vec<Argument>,
     pub args: Vec<FnArg<'a>>,
     pub output: syn::Type,
-    pub doc: syn::LitStr,
+    pub doc: PythonDoc,
     pub deprecations: Deprecations,
     pub convention: CallingConvention,
+    pub text_signature: Option<TextSignatureAttribute>,
 }
 
 pub fn get_return_info(output: &syn::ReturnType) -> syn::Type {
@@ -235,31 +258,22 @@ impl<'a> FnSpec<'a> {
         meth_attrs: &mut Vec<syn::Attribute>,
         options: PyFunctionOptions,
     ) -> Result<FnSpec<'a>> {
+        let PyFunctionOptions {
+            text_signature,
+            name,
+            mut deprecations,
+            ..
+        } = options;
+
         let MethodAttributes {
             ty: fn_type_attr,
             args: fn_attrs,
             mut python_name,
-        } = parse_method_attributes(meth_attrs, options.name.map(|name| name.0))?;
-
-        match fn_type_attr {
-            Some(MethodTypeAttribute::New) => {
-                if let Some(name) = &python_name {
-                    bail_spanned!(name.span() => "`name` not allowed with `#[new]`");
-                }
-                python_name = Some(syn::Ident::new("__new__", Span::call_site()))
-            }
-            Some(MethodTypeAttribute::Call) => {
-                if let Some(name) = &python_name {
-                    bail_spanned!(name.span() => "`name` not allowed with `#[call]`");
-                }
-                python_name = Some(syn::Ident::new("__call__", Span::call_site()))
-            }
-            _ => {}
-        }
+        } = parse_method_attributes(meth_attrs, name.map(|name| name.0), &mut deprecations)?;
 
         let (fn_type, skip_first_arg, fixed_convention) =
             Self::parse_fn_type(sig, fn_type_attr, &mut python_name)?;
-        Self::ensure_text_signature_on_valid_method(&fn_type, options.text_signature.as_ref())?;
+        Self::ensure_text_signature_on_valid_method(&fn_type, text_signature.as_ref())?;
 
         let name = &sig.ident;
         let ty = get_return_info(&sig.output);
@@ -267,11 +281,8 @@ impl<'a> FnSpec<'a> {
 
         let doc = utils::get_doc(
             meth_attrs,
-            options
-                .text_signature
-                .as_ref()
-                .map(|attr| (&python_name, attr)),
-        )?;
+            text_signature.as_ref().map(|attr| (&python_name, attr)),
+        );
 
         let arguments: Vec<_> = if skip_first_arg {
             sig.inputs
@@ -298,7 +309,8 @@ impl<'a> FnSpec<'a> {
             args: arguments,
             output: ty,
             doc,
-            deprecations: options.deprecations,
+            deprecations,
+            text_signature,
         })
     }
 
@@ -317,10 +329,7 @@ impl<'a> FnSpec<'a> {
                     "text_signature not allowed on __new__; if you want to add a signature on \
                      __new__, put it on the struct definition instead"
                 ),
-                FnType::FnCall(_)
-                | FnType::Getter(_)
-                | FnType::Setter(_)
-                | FnType::ClassAttribute => bail_spanned!(
+                FnType::Getter(_) | FnType::Setter(_) | FnType::ClassAttribute => bail_spanned!(
                     text_signature.kw.span() => "text_signature not allowed with this method type"
                 ),
                 _ => {}
@@ -364,14 +373,13 @@ impl<'a> FnSpec<'a> {
                 (FnType::ClassAttribute, false, None)
             }
             Some(MethodTypeAttribute::New) => {
+                if let Some(name) = &python_name {
+                    bail_spanned!(name.span() => "`name` not allowed with `#[new]`");
+                }
+                *python_name = Some(syn::Ident::new("__new__", Span::call_site()));
                 (FnType::FnNew, false, Some(CallingConvention::TpNew))
             }
             Some(MethodTypeAttribute::ClassMethod) => (FnType::FnClass, true, None),
-            Some(MethodTypeAttribute::Call) => (
-                FnType::FnCall(parse_receiver("expected receiver for #[call]")?),
-                true,
-                Some(CallingConvention::Varargs),
-            ),
             Some(MethodTypeAttribute::Getter) => {
                 // Strip off "get_" prefix if needed
                 if python_name.is_none() {
@@ -424,6 +432,17 @@ impl<'a> FnSpec<'a> {
         None
     }
 
+    pub fn is_pos_only(&self, name: &syn::Ident) -> bool {
+        for s in self.attrs.iter() {
+            if let Argument::PosOnlyArg(path, _) = s {
+                if path.is_ident(name) {
+                    return true;
+                }
+            }
+        }
+        false
+    }
+
     pub fn is_kw_only(&self, name: &syn::Ident) -> bool {
         for s in self.attrs.iter() {
             if let Argument::Kwarg(path, _) = s {
@@ -442,7 +461,7 @@ impl<'a> FnSpec<'a> {
         cls: Option<&syn::Type>,
     ) -> Result<TokenStream> {
         let deprecations = &self.deprecations;
-        let self_conversion = self.tp.self_conversion(cls);
+        let self_conversion = self.tp.self_conversion(cls, ExtractErrorMode::Raise);
         let self_arg = self.tp.self_arg();
         let arg_names = (0..self.args.len())
             .map(|pos| syn::Ident::new(&format!("arg{}", pos), Span::call_site()))
@@ -455,17 +474,17 @@ impl<'a> FnSpec<'a> {
             quote!(#func_name)
         };
         let rust_call =
-            quote! { pyo3::callback::convert(#py, #rust_name(#self_arg #(#arg_names),*)) };
+            quote! { ::pyo3::callback::convert(#py, #rust_name(#self_arg #(#arg_names),*)) };
         Ok(match self.convention {
             CallingConvention::Noargs => {
                 quote! {
                     unsafe extern "C" fn #ident (
-                        _slf: *mut pyo3::ffi::PyObject,
-                        _args: *mut pyo3::ffi::PyObject,
-                    ) -> *mut pyo3::ffi::PyObject
+                        _slf: *mut ::pyo3::ffi::PyObject,
+                        _args: *mut ::pyo3::ffi::PyObject,
+                    ) -> *mut ::pyo3::ffi::PyObject
                     {
                         #deprecations
-                        pyo3::callback::handle_panic(|#py| {
+                        ::pyo3::callback::handle_panic(|#py| {
                             #self_conversion
                             #rust_call
                         })
@@ -476,23 +495,23 @@ impl<'a> FnSpec<'a> {
                 let arg_convert_and_rust_call = impl_arg_params(self, cls, rust_call, &py, true)?;
                 quote! {
                     unsafe extern "C" fn #ident (
-                        _slf: *mut pyo3::ffi::PyObject,
-                        _args: *const *mut pyo3::ffi::PyObject,
-                        _nargs: pyo3::ffi::Py_ssize_t,
-                        _kwnames: *mut pyo3::ffi::PyObject) -> *mut pyo3::ffi::PyObject
+                        _slf: *mut ::pyo3::ffi::PyObject,
+                        _args: *const *mut ::pyo3::ffi::PyObject,
+                        _nargs: ::pyo3::ffi::Py_ssize_t,
+                        _kwnames: *mut ::pyo3::ffi::PyObject) -> *mut ::pyo3::ffi::PyObject
                     {
                         #deprecations
-                        pyo3::callback::handle_panic(|#py| {
+                        ::pyo3::callback::handle_panic(|#py| {
                             #self_conversion
-                            let _kwnames: Option<&pyo3::types::PyTuple> = #py.from_borrowed_ptr_or_opt(_kwnames);
+                            let _kwnames: ::std::option::Option<&::pyo3::types::PyTuple> = #py.from_borrowed_ptr_or_opt(_kwnames);
                             // Safety: &PyAny has the same memory layout as `*mut ffi::PyObject`
-                            let _args = _args as *const &pyo3::PyAny;
-                            let _kwargs = if let Some(kwnames) = _kwnames {
-                                std::slice::from_raw_parts(_args.offset(_nargs), kwnames.len())
+                            let _args = _args as *const &::pyo3::PyAny;
+                            let _kwargs = if let ::std::option::Option::Some(kwnames) = _kwnames {
+                                ::std::slice::from_raw_parts(_args.offset(_nargs), kwnames.len())
                             } else {
                                 &[]
                             };
-                            let _args = std::slice::from_raw_parts(_args, _nargs as usize);
+                            let _args = ::std::slice::from_raw_parts(_args, _nargs as usize);
 
                             #arg_convert_and_rust_call
                         })
@@ -503,15 +522,15 @@ impl<'a> FnSpec<'a> {
                 let arg_convert_and_rust_call = impl_arg_params(self, cls, rust_call, &py, false)?;
                 quote! {
                     unsafe extern "C" fn #ident (
-                        _slf: *mut pyo3::ffi::PyObject,
-                        _args: *mut pyo3::ffi::PyObject,
-                        _kwargs: *mut pyo3::ffi::PyObject) -> *mut pyo3::ffi::PyObject
+                        _slf: *mut ::pyo3::ffi::PyObject,
+                        _args: *mut ::pyo3::ffi::PyObject,
+                        _kwargs: *mut ::pyo3::ffi::PyObject) -> *mut ::pyo3::ffi::PyObject
                     {
                         #deprecations
-                        pyo3::callback::handle_panic(|#py| {
+                        ::pyo3::callback::handle_panic(|#py| {
                             #self_conversion
-                            let _args = #py.from_borrowed_ptr::<pyo3::types::PyTuple>(_args);
-                            let _kwargs: Option<&pyo3::types::PyDict> = #py.from_borrowed_ptr_or_opt(_kwargs);
+                            let _args = #py.from_borrowed_ptr::<::pyo3::types::PyTuple>(_args);
+                            let _kwargs: ::std::option::Option<&::pyo3::types::PyDict> = #py.from_borrowed_ptr_or_opt(_kwargs);
 
                             #arg_convert_and_rust_call
                         })
@@ -523,20 +542,20 @@ impl<'a> FnSpec<'a> {
                 let arg_convert_and_rust_call = impl_arg_params(self, cls, rust_call, &py, false)?;
                 quote! {
                     unsafe extern "C" fn #ident (
-                        subtype: *mut pyo3::ffi::PyTypeObject,
-                        _args: *mut pyo3::ffi::PyObject,
-                        _kwargs: *mut pyo3::ffi::PyObject) -> *mut pyo3::ffi::PyObject
+                        subtype: *mut ::pyo3::ffi::PyTypeObject,
+                        _args: *mut ::pyo3::ffi::PyObject,
+                        _kwargs: *mut ::pyo3::ffi::PyObject) -> *mut ::pyo3::ffi::PyObject
                     {
                         #deprecations
-                        use pyo3::callback::IntoPyCallbackOutput;
-                        pyo3::callback::handle_panic(|#py| {
-                            let _args = #py.from_borrowed_ptr::<pyo3::types::PyTuple>(_args);
-                            let _kwargs: Option<&pyo3::types::PyDict> = #py.from_borrowed_ptr_or_opt(_kwargs);
+                        use ::pyo3::callback::IntoPyCallbackOutput;
+                        ::pyo3::callback::handle_panic(|#py| {
+                            let _args = #py.from_borrowed_ptr::<::pyo3::types::PyTuple>(_args);
+                            let _kwargs: ::std::option::Option<&::pyo3::types::PyDict> = #py.from_borrowed_ptr_or_opt(_kwargs);
 
                             let result = #arg_convert_and_rust_call;
-                            let initializer: pyo3::PyClassInitializer::<#cls> = result.convert(#py)?;
+                            let initializer: ::pyo3::PyClassInitializer::<#cls> = result.convert(#py)?;
                             let cell = initializer.create_cell_from_subtype(#py, subtype)?;
-                            Ok(cell as *mut pyo3::ffi::PyObject)
+                            ::std::result::Result::Ok(cell as *mut ::pyo3::ffi::PyObject)
                         })
                     }
                 }
@@ -551,23 +570,23 @@ impl<'a> FnSpec<'a> {
         let doc = &self.doc;
         match self.convention {
             CallingConvention::Noargs => quote! {
-                pyo3::class::methods::PyMethodDef::noargs(
+                ::pyo3::class::methods::PyMethodDef::noargs(
                     #python_name,
-                    pyo3::class::methods::PyCFunction(#wrapper),
+                    ::pyo3::class::methods::PyCFunction(#wrapper),
                     #doc,
                 )
             },
             CallingConvention::Fastcall => quote! {
-                pyo3::class::methods::PyMethodDef::fastcall_cfunction_with_keywords(
+                ::pyo3::class::methods::PyMethodDef::fastcall_cfunction_with_keywords(
                     #python_name,
-                    pyo3::class::methods::PyCFunctionFastWithKeywords(#wrapper),
+                    ::pyo3::class::methods::PyCFunctionFastWithKeywords(#wrapper),
                     #doc,
                 )
             },
             CallingConvention::Varargs => quote! {
-                pyo3::class::methods::PyMethodDef::cfunction_with_keywords(
+                ::pyo3::class::methods::PyMethodDef::cfunction_with_keywords(
                     #python_name,
-                    pyo3::class::methods::PyCFunctionWithKeywords(#wrapper),
+                    ::pyo3::class::methods::PyCFunctionWithKeywords(#wrapper),
                     #doc,
                 )
             },
@@ -586,6 +605,7 @@ struct MethodAttributes {
 fn parse_method_attributes(
     attrs: &mut Vec<syn::Attribute>,
     mut python_name: Option<syn::Ident>,
+    deprecations: &mut Deprecations,
 ) -> Result<MethodAttributes> {
     let mut new_attrs = Vec::new();
     let mut args = Vec::new();
@@ -601,14 +621,19 @@ fn parse_method_attributes(
     }
 
     for attr in attrs.drain(..) {
-        match attr.parse_meta()? {
-            syn::Meta::Path(name) => {
+        match attr.parse_meta() {
+            Ok(syn::Meta::Path(name)) => {
                 if name.is_ident("new") || name.is_ident("__new__") {
                     set_ty!(MethodTypeAttribute::New, name);
                 } else if name.is_ident("init") || name.is_ident("__init__") {
                     bail_spanned!(name.span() => "#[init] is disabled since PyO3 0.9.0");
                 } else if name.is_ident("call") || name.is_ident("__call__") {
-                    set_ty!(MethodTypeAttribute::Call, name);
+                    deprecations.push(Deprecation::CallAttribute, name.span());
+                    ensure_spanned!(
+                        python_name.is_none(),
+                        python_name.span() => "`name` may not be used with `#[call]`"
+                    );
+                    python_name = Some(syn::Ident::new("__call__", Span::call_site()));
                 } else if name.is_ident("classmethod") {
                     set_ty!(MethodTypeAttribute::ClassMethod, name);
                 } else if name.is_ident("staticmethod") {
@@ -630,15 +655,19 @@ fn parse_method_attributes(
                     new_attrs.push(attr)
                 }
             }
-            syn::Meta::List(syn::MetaList {
+            Ok(syn::Meta::List(syn::MetaList {
                 path, mut nested, ..
-            }) => {
+            })) => {
                 if path.is_ident("new") {
                     set_ty!(MethodTypeAttribute::New, path);
                 } else if path.is_ident("init") {
                     bail_spanned!(path.span() => "#[init] is disabled since PyO3 0.9.0");
                 } else if path.is_ident("call") {
-                    set_ty!(MethodTypeAttribute::Call, path);
+                    ensure_spanned!(
+                        python_name.is_none(),
+                        python_name.span() => "`name` may not be used with `#[call]`"
+                    );
+                    python_name = Some(syn::Ident::new("__call__", Span::call_site()));
                 } else if path.is_ident("setter") || path.is_ident("getter") {
                     if let syn::AttrStyle::Inner(_) = attr.style {
                         bail_spanned!(
@@ -688,7 +717,7 @@ fn parse_method_attributes(
                     new_attrs.push(attr)
                 }
             }
-            syn::Meta::NameValue(_) => new_attrs.push(attr),
+            Ok(syn::Meta::NameValue(_)) | Err(_) => new_attrs.push(attr),
         }
     }
 
